import java.util.Map;
import java.util.Vector;

public interface Cmd extends Genable
//to implement CompliableProc it really needs to contain no references to process parameters
//Edit: No longer compilable, as a command doesnt gen its local variable info
{
	public Cmd substituteConst(Vector<String> vars, Vector<IConst> vals);
	public boolean usesSig(String sig, Map<String, Integer> consts);
}

class Skip implements Cmd
{
	public String genCode(Context con)
	{
		return "Skip";
	}

	public boolean usesSig(String sig, Map<String, Integer> consts){return false;}
	
	public String toString(){ return "skip";}


	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof Skip) ) return false;
		
		Skip that = (Skip)aThat;

		return true;
	}
	@Override public int hashCode() { 
		return 1 + 1;
	}
	
	
	public Cmd substituteConst(Vector<String> vars, Vector<IConst> vals) { return this;}

}


class Seq implements Cmd
{
	Cmd c1, c2;
	
	
	public Seq(Cmd c1, Cmd c2) 
	{

		if (c1==null) throw new ParserException("c1 Argument to Seq null");
		if (c2==null) throw new ParserException("c2 Argument to Seq null");		
		this.c1 = c1; this.c2=c2;

	}

	public Cmd substituteConst(Vector<String> vars, Vector<IConst> vals) { return new Seq(c1.substituteConst(vars,vals),c2.substituteConst(vars,vals));}
	public String genCode(Context con)
	{
		return "Sq.(" + c1.genCode(con) + "," + c2.genCode(con) + ")";
	}

	public boolean usesSig(String sig, Map<String, Integer> consts){return c1.usesSig(sig,consts) || c2.usesSig(sig,consts);}

	
	public String toString()
	{ return c1 + "; " + c2;}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof Seq) ) return false;
		
		Seq that = (Seq)aThat;

		return ( this.c1.equals(that.c1)
			&& this.c2.equals(that.c2));
	}
	@Override public int hashCode() { 
		return 1 + c1.hashCode()
			+ c2.hashCode();
	}
	
}

class Iter implements Cmd
{
	Cmd c;
	public Iter(Cmd c) 
	{
		if (c==null) throw new ParserException("Argument to Iter null");	
		this.c = c;
	}
	
	public Cmd substituteConst(Vector<String> vars, Vector<IConst> vals) { return new Iter(c.substituteConst(vars,vals));}
	public String genCode(Context con)
	{
		return "Iter." + c.genCode(con);
	}
	
	public boolean usesSig(String sig, Map<String, Integer> consts){return c.usesSig(sig,consts);}
	public String toString() {return "iter { " + c + " }";} 
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof Iter) ) return false;
		
		Iter that = (Iter)aThat;

		return ( this.c.equals(that.c));
	}
	@Override public int hashCode() { 
		return 1 + c.hashCode();
	}
}

class While implements Cmd
{
	Cmd c;
	BExpr b;
	public While(BExpr b, Cmd c) 
	{
		if (b==null) throw new ParserException("b argument to While null");
		if (c==null) throw new ParserException("c argument to While null");
		this.b=b; this.c = c;
	}
	
	public Cmd substituteConst(Vector<String> vars, Vector<IConst> vals) { return new While(b.substituteConst(vars,vals),c.substituteConst(vars,vals));}
	public String genCode(Context con)
	{
		return "While.(" + b.genCode(con) + "," + c.genCode(con) + ")";
	}
	
	
	public boolean usesSig(String sig, Map<String, Integer> consts){return c.usesSig(sig,consts);}

	public String toString() { return "while " + b + " do { " + c + " }";}
	
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof While) ) return false;
		
		While that = (While)aThat;

		return ( this.c.equals(that.c)
			&& this.b.equals(that.b));
	}
	@Override public int hashCode() { 
		return 1 + c.hashCode()
			+ b.hashCode();
	}
}


class If implements Cmd
{
	Cmd c1, c2;
	BExpr b;
	public If(BExpr b, Cmd c1, Cmd c2) 
	{
		if (b==null) throw new ParserException("b argument to If null");
		if (c1==null) throw new ParserException("c1 argument to If null");
		if (c2==null) throw new ParserException("c2 argument to If null");
		
		this.b=b; this.c1 = c1; this.c2 = c2;

	}
	public If(BExpr b, Cmd c1) {this(b,c1,new Skip());}
	
	public Cmd substituteConst(Vector<String> vars, Vector<IConst> vals) 
	{ return new If(b.substituteConst(vars,vals),c1.substituteConst(vars,vals),c2.substituteConst(vars,vals));}
	public String genCode(Context con)
	{
		return "Cond.(" + b.genCode(con) + "," + c1.genCode(con) + "," + c2.genCode(con) + ")";
	}
	
	public boolean usesSig(String sig, Map<String, Integer> consts){return c1.usesSig(sig,consts) || c2.usesSig(sig,consts);}

	
	public String toString() { return "if " + b + " then { " + c1 + " } else { " + c2 + " }"; }
	
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof If) ) return false;
		
		If that = (If)aThat;

		return ( this.c1.equals(that.c1)
			&& this.c2.equals(that.c2)
			&& this.b.equals(that.b));
	}
	@Override public int hashCode() { 
		return 1 + c1.hashCode()
			+ c2.hashCode()
			+ b.hashCode();
	}
}

class IAssign implements Cmd
{
	IExpr iexpr1,iexpr2;
	public IAssign(IExpr iexpr1, IExpr iexpr2) 
	{
		if (iexpr1==null) throw new ParserException("iexpr1 argument to IAssign null");
		if (iexpr2==null) throw new ParserException("iexpr2 argument to IAssign null");
		this.iexpr1 = iexpr1; this.iexpr2 = iexpr2;
	}
	
	public Cmd substituteConst(Vector<String> vars, Vector<IConst> vals) 
	{ return new IAssign(iexpr1.substituteConst(vars,vals),iexpr2.substituteConst(vars,vals));}
	public String genCode(Context con)
	{
		return "Iassign.(" + iexpr1.genCode(con) + "," + iexpr2.genCode(con) + ")";
	}
	
	
	public boolean usesSig(String sig, Map<String, Integer> consts){return false;}

	
	public String toString() { return iexpr1 + ":= " + iexpr2;}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof IAssign) ) return false;
		
		IAssign that = (IAssign)aThat;

		return ( this.iexpr1.equals(that.iexpr1)
			&& this.iexpr2.equals(that.iexpr2));
	}
	@Override public int hashCode() { 
		return 1 + iexpr1.hashCode()
			+ iexpr2.hashCode();
	}
}	

	
class BAssign implements Cmd
{
	BExpr bexpr1, bexpr2;
	public BAssign(BExpr bexpr1, BExpr bexpr2) 
	{
		if (bexpr1==null) throw new ParserException("bvname argument to BAssign null");
		if (bexpr2==null) throw new ParserException("bexpr argument to BAssign null");
		 
		this.bexpr1 = bexpr1; this.bexpr2 = bexpr2;
	}
	
	public Cmd substituteConst(Vector<String> vars, Vector<IConst> vals) 
	{ return new BAssign(bexpr1.substituteConst(vars,vals),bexpr2.substituteConst(vars,vals));}
	
	public String genCode(Context con)
	{
		return "Bassign.(" + bexpr1.genCode(con) + "," + bexpr2.genCode(con) + ")";
	}
	
	public boolean usesSig(String sig, Map<String, Integer> consts){return false;}

	
	public String toString() {return bexpr1 + ":= " + bexpr2;}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof BAssign) ) return false;
		
		BAssign that = (BAssign)aThat;

		return ( this.bexpr1.equals(that.bexpr1)
			&& this.bexpr2.equals(that.bexpr2));
	}
	@Override public int hashCode() { 
		return 1 + bexpr1.hashCode()
			+ bexpr2.hashCode();
	}
}	

	
class Sig implements Cmd
{
	String s;
	Vector<IConst> indices;
	public Sig(String s,Vector<IConst> indices) 
	{
		if (s==null) throw new ParserException("s argument to Sig null");
		if (indices==null) throw new ParserException("indices argument to Sig null");
		this.s = s;
		this.indices = indices;
	}
	

	public Cmd substituteConst(Vector<String> vars, Vector<IConst> vals) 
	{ 
		Vector<IConst> newIndices = new Vector<IConst>();
		for (IConst i : indices)
			newIndices.add(i.substituteConst(vars,vals));
		return new Sig(s,newIndices);
	}
	
	
	public String genCode(Context con)
	{
		String result = "Sig." + s;
		for (IConst i : indices)
		{ result += "." + i.genCode(con);}
		return result;
	}
	
	public boolean usesSig(String sig, Map<String, Integer> consts)
	{
		String thisSig = s;
		for (IConst i : indices)
		{ thisSig += "." + i.constVal(consts);}
		return thisSig.equals(sig);
	}

	
	public String toString() 
	{ 
		String result =  "sig(" + s;
		for (IConst i : indices)
		{ result += "." + i.toString();}
		return result + ")";

	}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof Sig) ) return false;
		
		Sig that = (Sig)aThat;

		return ( this.s.equals(that.s)
			&& this.indices.equals(that.indices));
	}
	@Override public int hashCode() { 
		return 1 + s.hashCode()
			+ indices.hashCode();
	}
}
	
class ISig implements Cmd
{
	String s;
	IExpr i;
	Vector<IConst> indices;
	public ISig(String s, Vector<IConst> indices, IExpr i) 
	{
		if (s==null) throw new ParserException("s argument to ISig null");
		if (i==null) throw new ParserException("i argument to ISig null");
		if (indices==null) throw new ParserException("indices argument to ISig null");
	
		this.s = s; this.i=i; this.indices = indices;
	}
	
	public Cmd substituteConst(Vector<String> vars, Vector<IConst> vals) 
	{ 
		Vector<IConst> newIndices = new Vector<IConst>();
		for (IConst j : indices)
			newIndices.add(j.substituteConst(vars,vals));
		return new ISig(s,newIndices,i.substituteConst(vars,vals) );
	}
	
	public String genCode(Context con)
	{
		String result = "ISig.(" + s;
		for (IConst ic : indices)
		{ result += "." + ic.genCode(con);}
		return result + "," + i.genCode(con) + ")";
	}
	
	
	public boolean usesSig(String sig, Map<String, Integer> consts)
	{
		String thisSig = s;
		for (IConst i : indices)
		{ thisSig += "." + i.constVal(consts);}
		return thisSig.equals(sig);
	}
	
	
	public String toString() 
	{
		String result =  "isig(" + s;
		for (IConst ic : indices)
		{ result += "." + ic.toString();}
		return result + ", " + i + ")";
	}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof ISig) ) return false;
		
		ISig that = (ISig)aThat;

		return ( this.s.equals(that.s)
			&& this.i.equals(that.i)
			&& this.indices.equals(that.indices));
	}
	@Override public int hashCode() { 
		return 1 + s.hashCode()
			+ i.hashCode()
			+ indices.hashCode();
	}
}

class Atomic implements Cmd
{
	Cmd c;
	public Atomic(Cmd c) 
	{
		if (c==null) throw new ParserException("Argument to Atomic null");
		this.c = c;
	}
	
	public Cmd substituteConst(Vector<String> vars, Vector<IConst> vals) { return new Atomic(c.substituteConst(vars,vals));}
	
	public String genCode(Context con)
	{
		return "Atomic." + c.genCode(con);
	}
	
	public boolean usesSig(String sig, Map<String, Integer> consts){return c.usesSig(sig,consts);}
	
	public String toString() {return "atomic " + c;}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof Atomic) ) return false;
		
		Atomic that = (Atomic)aThat;

		return ( this.c.equals(that.c));
	}
	@Override public int hashCode() { 
		return 1 + c.hashCode();
	}
}












