//Structures to match the CSPm structures wit the same name 
//datatype CmdStruct = CSLeaf.Cmd | CSNode.Seq(CmdStruct) 
//         | CSMonitor.Cmd

import java.util.*;

public interface CmdStruct extends ProcessContainer
{
	public CmdStruct clone();
}

class CSLeaf implements CmdStruct
{
	CompilableProc p;
	
	public CSLeaf(CompilableProc p) {this.p = p;}

	public int numProcs() {return 1;}
	public String process(int i){if (i == 0) return p.toString(); else throw new InterpreterException("CSLeaf asked for process greater than 1 (" + i + ")");}
	public String ownerOfSig(String sig,Map<String,Integer> consts){if (p.usesSig(sig,consts)) return p.toString(); else throw new InterpreterException("CSLeaf asked for owner of unused sig");}
	public int indexOfOwnerOfSig(String sig,Map<String,Integer> consts){if (p.usesSig(sig,consts)) return 0; else throw new InterpreterException("CSLeaf asked for index of owner of unused sig");}
	public boolean usesSig(String sig, Map<String, Integer> consts){return p.usesSig(sig,consts);}
	public String genCode(Context con)
	{
		return "CSLeaf." + p.genCode(con);
	}
	//Keep the same processes but clone the (mutable) storage (ie the vectors) in CSNode
	public CmdStruct clone(){return new CSLeaf(p);}
	
	public Iterator<CompilableProc> iterator()
	{
		return new Iterator<CompilableProc>()
		{
			public boolean hasNext = true;
			public boolean hasNext(){return hasNext;}
			public CompilableProc next()
			{ 
				if (hasNext)
				{
					hasNext = false;
					return p;
				}
				else throw new NoSuchElementException();
			}
			public void remove() {throw new UnsupportedOperationException();}
		};
	}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof CSLeaf) ) return false;
		
		CSLeaf that = (CSLeaf)aThat;

		return this.p.equals(that.p);
	}
	@Override public int hashCode() { 
		return 1 + p.hashCode();
	}

	@Override public String toString() {
		return p.toString();
	
	}
}

class CSOverseer extends CSLeaf
{	
	
	public CSOverseer(CompilableProc p) {super(p);}
	public String genCode(Context con)
	{
		return "CSOverseer." + p.genCode(con);
	}
	public CmdStruct clone(){return new CSOverseer(p);}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof CSOverseer) ) return false;
		
		CSOverseer that = (CSOverseer)aThat;

		return this.p.equals(that.p);
	}
	@Override public int hashCode() { 
		return 1 + p.hashCode();
	}
	@Override public String toString() {
		return "overseer " + p.toString();
	
	}
}

class CSNode implements CmdStruct
{
	Vector<CmdStruct> seq;
	public CSNode(Vector<CmdStruct> seq){this.seq = seq;}

	public int numProcs()
	{
		int i = 0;
		for (CmdStruct c : seq)
		{
			i += c.numProcs();
		}
		return i;
	}
	
	public String process(int i)
	{
		int j = i;
		for (CmdStruct c : seq)
		{
			if (j < c.numProcs())
				return c.process(j);
			else
				j -= c.numProcs();
		}
		throw new InterpreterException("CSNode.process - index out of bounds");
	}
	public String ownerOfSig(String sig,Map<String,Integer> consts)
	{
		for (CmdStruct c : seq)
		{
			if (c.usesSig(sig,consts)) return c.ownerOfSig(sig,consts);
		}
		throw new InterpreterException("CSNode asked for owner of sig it doenst use");
	}
	public int indexOfOwnerOfSig(String sig,Map<String,Integer> consts)
	{
		int i = 0;
		for (CmdStruct c : seq)
		{
			if (c.usesSig(sig,consts)) return i + c.indexOfOwnerOfSig(sig,consts);
			else i += c.numProcs();
		}
		throw new InterpreterException("CSNode asked for index of owner of sig it doenst use");
	}
	public boolean usesSig(String sig, Map<String, Integer> consts)
	{
		for (CmdStruct c : seq)
		{
			if (c.usesSig(sig,consts)) return true;
		}
		return false;
	}
	public String genCode(Context con)
	{
		String s = "CSNode.<";
		for (CmdStruct c : seq)
		{
			s += c.genCode(con) + ",";
		}
		if (s.endsWith(",")) s = s.substring(0, s.length() -1);
		s += ">";
		return s;
	}
	
	public CmdStruct clone()
	{
		Vector<CmdStruct> v= new Vector<CmdStruct>();
		for (CmdStruct c : seq)
		{
			v.add(c.clone());
		}
		return new CSNode(v);
	}
	
	public Iterator<CompilableProc> iterator()
	{
		return new Iterator<CompilableProc>()
		{
			Iterator<CmdStruct> cmdStructs = seq.iterator();
			Iterator<CompilableProc> current = (new Vector<CompilableProc>()).iterator();
		
			private void findNext()
			{
				while (!current.hasNext() && cmdStructs.hasNext())
				{
					current = cmdStructs.next().iterator();
				}
			
			}
		
			public boolean hasNext()
			{
				findNext();
				return current.hasNext();
			}
			public CompilableProc next()
			{ 
				findNext();
				return current.next();
			}
			public void remove() {throw new UnsupportedOperationException();}
		};
	}
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof CSNode) ) return false;
		
		CSNode that = (CSNode)aThat;

		return this.seq.equals(that.seq);
	}
	
	@Override public int hashCode() { 
		return 1 + seq.hashCode();
	}
	
	@Override public String toString() {
		
		String result = "<";
		for (CmdStruct c : seq)
			result += c.toString() + ",";
			
		if (result.endsWith(",")) 
			result = result.substring(0, result.length() -1);
		return result + ">";
	
	}
}

//Bit of a hack to allow paramaterised programs
interface PreCmdStruct
{
	CmdStruct generate(Compiler compiler, Map<String, Integer> argVals);
}

interface ProcessParameter
{
	int value (Compiler compiler, Map<String,Integer> argVals);
}

class PreCSLeaf implements PreCmdStruct
{
	String name;
	Vector<ProcessParameter> args;
	
	public PreCSLeaf(String name, Vector<ProcessParameter> args){this.name = name; this.args= args;}
	
	public CmdStruct generate(Compiler compiler, Map<String, Integer> argVals)
	{
		Process process = compiler.procs.get(name);
		if (process.numArgs() != args.size()) 
			throw new ParserException(name + " given " + args.size() + " arguments but expects " + process.numArgs());
		
		Vector<Integer> argNums = new Vector<Integer>();
		
		for (ProcessParameter a : args)
		{
			argNums.add(a.value(compiler,argVals));
		}
		
		
		return new CSLeaf( new InitializedProc(process, argNums));
	}

}
class PreCSOverseer implements PreCmdStruct
{
	String name;
	Vector<ProcessParameter> args;
	
	public PreCSOverseer(String name, Vector<ProcessParameter> args){this.name = name; this.args= args;}
	
	public CmdStruct generate(Compiler compiler, Map<String, Integer> argVals)
	{
		Process process = compiler.procs.get(name);
		if (process.numArgs() != args.size()) 
			throw new ParserException(name + " given " + args.size() + " arguments but expects " + process.numArgs());
		
		Vector<Integer> argNums = new Vector<Integer>();
		
		for (ProcessParameter a : args)
		{
			argNums.add(a.value(compiler,argVals));
		}
		
		
		return new CSOverseer( new InitializedProc(process, argNums));
	}
}

class PreCSNode implements PreCmdStruct
{
	Vector<PreCmdStruct> seq;
	public PreCSNode(Vector<PreCmdStruct> seq){ this.seq = seq;}
	
	public CmdStruct generate(Compiler compiler, Map<String, Integer> argVals)
	{
		Vector<CmdStruct> newSeq = new Vector<CmdStruct>();
		
		for (PreCmdStruct p :  seq)
			newSeq.add(p.generate(compiler,argVals));
			
		return new CSNode(newSeq);
	}
}

class PreCSParamaterisedNode implements PreCmdStruct
{
	String var;
	ProcessParameter min, max;
	Vector<PreCmdStruct> seq;
	public PreCSParamaterisedNode(Vector<PreCmdStruct> seq, String var, ProcessParameter min, ProcessParameter max)
	{ 
		this.seq = seq; this.var = var; this.min = min; this.max = max;
	}
	
	
	public CmdStruct generate(Compiler compiler, Map<String, Integer> argVals)
	{
		Vector<CmdStruct> newSeq = new Vector<CmdStruct> ();
		
		for(int i = min.value(compiler,argVals); i <= max.value(compiler,argVals); i++)
		{
			argVals.put(var, i);
			for (PreCmdStruct p : seq)
				newSeq.add(p.generate(compiler,argVals));
		}
		
		return new CSNode(newSeq);
	
	}
}

class NumParam implements ProcessParameter
{
	int num;
	public NumParam(int num){this.num = num;}

	public int value (Compiler compiler, Map<String,Integer> argVals)
	{return num;}
}

class ConstParam  implements ProcessParameter
{
	String name;
	public ConstParam(String name){this.name = name;}

	public int value (Compiler compiler, Map<String,Integer> argVals)
	{return compiler.context.getConsts().get(name);}
}

class NameParam implements  ProcessParameter
{
	String name;
	public NameParam(String name){this.name = name;}

	public int value (Compiler compiler, Map<String,Integer> argVals)
	{
		if (! argVals.containsKey(name))
			throw new ParserException("Identifier " + name +  " undefined");
			
		return argVals.get(name);
	}
}

class AddParam implements ProcessParameter
{
	ProcessParameter p1,p2;
	public AddParam(ProcessParameter p1,ProcessParameter p2){this.p1 = p1; this.p2 = p2;}

	public int value (Compiler compiler, Map<String,Integer> argVals)
	{ return p1.value(compiler,argVals) + p2.value(compiler,argVals);}
}

class SubParam implements ProcessParameter
{
	ProcessParameter p1,p2;
	public SubParam(ProcessParameter p1,ProcessParameter p2){this.p1 = p1; this.p2 = p2;}

	public int value (Compiler compiler, Map<String,Integer> argVals)
	{ return p1.value(compiler,argVals) - p2.value(compiler,argVals);}
}

class MultParam implements ProcessParameter
{
	ProcessParameter p1,p2;
	public MultParam(ProcessParameter p1,ProcessParameter p2){this.p1 = p1; this.p2 = p2;}

	public int value (Compiler compiler, Map<String,Integer> argVals)
	{ return p1.value(compiler,argVals) * p2.value(compiler,argVals);}
}

class DivParam implements ProcessParameter
{
	ProcessParameter p1,p2;
	public DivParam(ProcessParameter p1,ProcessParameter p2){this.p1 = p1; this.p2 = p2;}

	public int value (Compiler compiler, Map<String,Integer> argVals)
	{ return p1.value(compiler,argVals) / p2.value(compiler,argVals);}
}

class ModParam implements ProcessParameter
{
	ProcessParameter p1,p2;
	public ModParam(ProcessParameter p1,ProcessParameter p2){this.p1 = p1; this.p2 = p2;}

	public int value (Compiler compiler, Map<String,Integer> argVals)
	{ return p1.value(compiler,argVals) % p2.value(compiler,argVals);}
}

class NegParam implements ProcessParameter
{
	ProcessParameter p1;
	public NegParam(ProcessParameter p1){this.p1 = p1;}

	public int value (Compiler compiler, Map<String,Integer> argVals)
	{ return - p1.value(compiler,argVals);}
}
