import java.util.Vector;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;


public interface Program extends ProcessContainer 
{
	public Program unnamedCopy();
	public String getName(); //returns null if program is unnamed
	public void addProcess(CompilableProc proc); //adds a process to the program at some location - intended to be used by assertions

	public String genCode(Context con, String visible); //additional parameter will be passed as second parameter to compile - ie it should tell with events should be left visible
}

class CmdStructProgram  implements Program
{

	CmdStruct CS;
	String name;
	String compilationCommand;//To allow for the various different possibilities

	
	public CmdStructProgram(CmdStruct CS)
	{
		this(null,CS,"StructuredCompile");
	}
	public CmdStructProgram(String name, CmdStruct CS)
	{
		this(name,CS,"StructuredCompile");
	}	
	public CmdStructProgram(CmdStruct CS, String compilationCommand)
	{
		this(null,CS,compilationCommand);
	}

	public CmdStructProgram(String name, CmdStruct CS, String compilationCommand)
	{
		if (CS == null) throw new ParserException("CS argument to CmdStructProgram null");
		this.CS = CS;
		this.name = name;
		this.compilationCommand = compilationCommand;
	}
	
	public String genCode(Context con)
	{
		 //return genCode(con, "(Union({Signals,ISignals,{|error,verror|}}))");
		 return genCode(con, "({| c | c <-Union({Signals,ISignals',Errors}) |})");
	}
	
	public String genCode(Context con, String visible)
	{
		String s = compilationCommand + "(" + CS.genCode(con) + ") " + visible ;
		if (name == null)
			return s;
		else return name + " = " + s;
	}
	
	public int numProcs(){return CS.numProcs();}
	public String process(int i) {return CS.process(i);}
	public String ownerOfSig(String sig,Map<String,Integer> consts){return CS.ownerOfSig(sig,consts);}
	public int indexOfOwnerOfSig(String sig,Map<String,Integer> consts){ return CS.indexOfOwnerOfSig(sig,consts);}
	public boolean usesSig(String sig, Map<String, Integer> consts)	{return CS.usesSig(sig,consts);}

	public Program unnamedCopy()
	{
		return new CmdStructProgram(CS.clone(),compilationCommand);
	}

	public String getName(){return name;}
	public void addProcess(CompilableProc cp)
	{
		Vector<CmdStruct> v = new Vector<CmdStruct>();
		v.add(CS);
		v.add(new CSLeaf(cp));

		CS = new CSNode(v);
	}
	
	public Iterator<CompilableProc> iterator(){return CS.iterator();}
	
	public String toString()
	{
		if (name != null)
			return name;
		else return compilationCommand + "(" + CS.toString() + ")";
	}
	
	
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof CmdStructProgram) ) return false;
		
		CmdStructProgram that = (CmdStructProgram)aThat;

		if (!(this.CS.equals(that.CS) && this.compilationCommand.equals(that.compilationCommand)))
			return false;
		if (this.name == null)
			return (that.name == null);
		if (that.name == null)
			return false;
		return this.name.equals(that.name);
	}
	@Override public int hashCode() { 
		return 1 + CS.hashCode() + (name == null ? 1 : name.hashCode()) + compilationCommand.hashCode();
	}
	
}


class InitializedProc implements CompilableProc
{

	Process proc;
	Vector<Integer> args;
	
	Map<String, Integer> argMap;
	
	
	
	public InitializedProc(Process proc, Vector<Integer>args)
	{
		if (proc == null) throw new ParserException("proc argument to InitializedProc null");
		if (args == null) throw new ParserException("args argument to InitializedProc null");
		if (args.size() != proc.consts.size()) throw new ParserException("InitializedProc: args.size() and proc.consts.size() do not match");
		Debug.trace("New Initilized Proc - proc: " + proc + " args: " + args);

		this.proc = proc; this.args=args;
	
		argMap = new TreeMap<String, Integer>();
		
		int i = 0;
		while (i < args.size())
		{
			argMap.put(proc.consts.elementAt(i), args.elementAt(i));
			i++;
		}
	}
	
	public String genCode(Context con){return genCode();}
	
	public String genCode()
	{
		String s = proc.name + "(";
			
		Iterator<Integer> j= args.iterator();
		while (j.hasNext())
		{
			Integer next = j.next();
			s = s + next;
			if (j.hasNext()) s = s + ",";
		}
		s = s + ")";
		return s;
	}

	public boolean usesSig(String sig, Map<String, Integer> consts)
	{
		Map<String,Integer> allConsts = new TreeMap<String,Integer>(consts);
		allConsts.putAll(argMap);
		return proc.usesSig(sig,allConsts);
	}
	
	public String toString()
	{ return this.genCode();}
	
	public Vector<LocalVarUsage> getLocalIVars()
	{
		Debug.trace("getLocalIVars() - proc.localIVars.size() = " + proc.getLocalIVars().size());
		Vector<LocalVarUsage> result = new Vector<LocalVarUsage>();
		for (String var: proc.getLocalIVars())
		{
			result.add(new LocalVarUsage(var, proc.name ,args));
		}
		return result;
	}	
	public Vector<LocalVarUsage> getLocalBVars()
	{
		Vector<LocalVarUsage> result = new Vector<LocalVarUsage>();
		for (String var: proc.getLocalBVars())
		{
			result.add(new LocalVarUsage(var, proc.name ,args));
		}
		return result;
	}	
	public Vector<LocalVarUsage> getLocalIAVars()
	{
		Vector<LocalVarUsage> result = new Vector<LocalVarUsage>();
		for (String var: proc.getLocalIAVars())
		{
			result.add(new LocalVarUsage(var, proc.name ,args));
		}
		return result;
	}	
	public Vector<LocalVarUsage> getLocalBAVars()
	{
		Vector<LocalVarUsage> result = new Vector<LocalVarUsage>();
		for (String var: proc.getLocalBAVars())
		{
			result.add(new LocalVarUsage(var, proc.name ,args));
		}
		return result;
	}
	
	
	@Override public boolean equals(Object aThat) 
	{
		//check for self-comparison
		if ( this == aThat ) return true;

		if ( !(aThat instanceof InitializedProc) ) return false;
		
		InitializedProc that = (InitializedProc)aThat;

		return (this.proc.equals(that.proc) && this.args.equals(that.args) && this.argMap.equals(that.argMap));
	}
	@Override public int hashCode() { 
		return 1 + proc.hashCode() + args.hashCode() + argMap.hashCode();
	}
}









